#!/usr/bin/env /usr/local/bin/node
import { program } from 'commander';
import debug from 'debug';
import dotenv from 'dotenv-extended';
import PgBoss from 'pg-boss';
import dev from './commandDev.js';
import run from './commandRun.js';
import enqueue from './enqueue.js';
import receive from './receive.js';
import withDatabaseClient from './withDatabaseClient.js';

dotenv.load();

const log = debug('catirpel');
debug.enable('catirpel*');

program.version('0.130.0');

program
  .command('dev <caterpillarPath> <message...>')
  .description('Develop a caterpillar')
  .action((caterpillarPath, message) => {
    dev(
      caterpillarPath,
      message
        .map(x => x.split('='))
        .map(([key, ...value]) => ({ [key]: value.join('=') }))
        .reduce((acc, x) => ({ ...acc, ...x }), {})
    );
  });

program
  .command('run <folders...>')
  .description('Run caterpillars')
  .action(folders => {
    run(folders);
  });

program
  .command('enqueue <caterpillar>')
  .description('Enqueue messages')
  .action(async caterpillar => {
    const message = receive();
    const messages = Array.isArray(message) ? message : [message];
    const boss = new PgBoss({});
    await boss.start();
    await Promise.all(messages.map(x => enqueue(boss, caterpillar, x)));
    process.exit(0);
  });

program
  .command('info')
  .description('Show queue statistics')
  .option('-j, --json', 'output json')
  .action(options =>
    withDatabaseClient(async client => {
      const queues = await client.query(`select name, state, count(*)::int
        from pgboss.job
        where name not like '\\_\\_%'
        group by name, state
        order by name`);
      const data = await client.query(`select
          relname as name,
          n_live_tup::int as count
        from pg_stat_user_tables
        where schemaname = 'public'
        order by n_live_tup desc`);
      const output = {
        queues: queues.rows,
        data: data.rows,
      };
      console.log(options.json ? JSON.stringify(output, undefined, 2) : output);
    })
  );

program
  .command('list <caterpillar> <state>')
  .description('List queued messages')
  .action((caterpillar, state) =>
    withDatabaseClient(async client => {
      const info = await client.query(
        `select distinct data
        from pgboss.job
        where name = $1 and state = $2
      `,
        [caterpillar, state]
      );
      info.rows.forEach(({ data: data }) =>
        console.log(data.url ?? data.id ?? data)
      );
    })
  );

program
  .command('reset [caterpillar]')
  .description('Delete queues')
  .action(async caterpillar => {
    const boss = new PgBoss({});
    await boss.start();
    if (caterpillar) {
      log('deleting queue for', caterpillar);
      await boss.deleteQueue(caterpillar);
    } else {
      log('clearing storage');
      await boss.clearStorage();
    }
    process.exit(0);
  });

program.parse();
