import { exec } from 'child_process';
import debug from 'debug';
import fs from 'fs';
import { join } from 'path';
import PgBoss from 'pg-boss';
import cleanDocument from './cleanDocument.js';
import enqueue from './enqueue.js';
import insertDocument from './insertDocument.js';
import withDatabaseClient from './withDatabaseClient.js';

const log = debug('catirpel:run');

const listCaterpillars = async folders =>
  (
    await Promise.all(
      folders.map(async root => {
        const files = await fs.promises.readdir(root);
        const caterpillars = [...new Set(files.map(x => x.split('.')[0]))];
        return caterpillars.map(caterpillar => ({
          name: caterpillar,
          path: join(root, `${caterpillar}.js`),
        }));
      })
    )
  ).flat(1);

const process = (caterpillar, data) => {
  return new Promise((resolve, reject) => {
    exec(
      `node ${caterpillar.path} '${JSON.stringify(data)}'`,
      {
        env: {},
        shell: true,
      },
      (error, stdout, stderr) => {
        if (error) {
          new Error('Processor error');
          reject(error);
        } else {
          if (stderr) {
            log(caterpillar.name, 'reports:', stderr);
          }
          resolve(stdout);
        }
      }
    );
  });
};

const run = async folders => {
  // Initialise caterpillars
  const caterpillars = await listCaterpillars(folders);
  const caterpillarMap = caterpillars.reduce((acc, x) => {
    acc[x.name] = x;
    return acc;
  }, {});
  log('found', caterpillars.length, 'caterpillars', caterpillars);

  // Initialise queue
  const boss = new PgBoss({});
  boss.on('error', error => {
    log('queue error', error);
  });
  await boss.start();
  await boss.onComplete('*', job => {
    if (job.data.state === 'failed') {
      log('abandoned message', job.data);
    }
  });
  log('initialised queue');

  // Create data tables
  await withDatabaseClient(client =>
    Promise.all(
      caterpillars.map(x =>
        client.query(
          `create table if not exists public.${x.name} (
            id text not null primary key,
            updated_at timestamp with time zone default now() not null,
            domain text,
            document jsonb not null
          )`
        )
      )
    )
  );
  log('created data tables');

  // Crawl
  await Promise.all(
    caterpillars.map(x =>
      boss.subscribe(x.name, { newJobCheckIntervalSeconds: 5 }, async job => {
        const caterpillar = caterpillarMap[job.name];
        log(caterpillar.name, 'received a message', job.data);
        try {
          const result = JSON.parse(await process(caterpillar, job.data));
          log(
            caterpillar.name,
            'got',
            result.messages.length,
            'messages and',
            result.documents.length,
            'documents'
          );

          // Enqueue messages
          await Promise.all(
            result.messages.map(message =>
              enqueue(boss, caterpillar.name, message)
            )
          );

          // Insert/update documents
          await withDatabaseClient(client =>
            Promise.all(
              result.documents
                .map(x => cleanDocument(x))
                .map(x => insertDocument(client, caterpillar.name, x))
            )
          );

          job.done();
        } catch (error) {
          log(caterpillar.name, 'fell into a hole', error);
          throw error;
        }
      })
    )
  );
};

export default run;
