import isDebug from './isDebug.js';
import resolveUrl from './resolveUrl.js';

const apply = (x, func) =>
  Array.isArray(x)
    ? Promise.all(x.map(func))
    : x !== undefined
    ? func(x)
    : undefined;

const has = x =>
  Array.isArray(x) ? x.length !== 0 : x !== undefined && x !== null;

const evaluate = async (page, obj) => {
  let results;
  try {
    if (obj.$) {
      results = await page.$(obj.$);
    } else if (obj.$$) {
      results = await page.$$(obj.$$);
    }
  } catch (error) {
    if (process.env.DEBUG) {
      if (isDebug) {
        console.error(error);
      }
    }
  }
  if (!results || results.length === 0) {
    return undefined;
  }

  // Cast
  if (obj.cast) {
    if (obj.cast === 'text') {
      results = await apply(results, async x => (await x.textContent()).trim());
    } else if (obj.cast === 'href') {
      results = await apply(results, async x =>
        apply(await x?.getAttribute('href'), y => resolveUrl(page, y))
      );
    } else if (obj.cast === 'int') {
      results = await apply(results, async x =>
        parseInt((await x?.textContent())?.replace(/[^\d]/g, ''), 10)
      );
    } else if (obj.cast === 'float') {
      results = await apply(results, async x =>
        parseFloat((await x?.textContent())?.replace(/[^\d.]/g, ''))
      );
    } else if (obj.cast === 'floatDe') {
      results = await apply(results, async x =>
        parseFloat(
          (await x?.textContent())?.replace(/[^\d,]/g, '')?.replace(/,/g, '.')
        )
      );
    }
  }

  // Map results
  if (obj.map) {
    results = await apply(results, obj.map);
  } else {
    results = await apply(results, async x =>
      x.textContent ? (await x?.textContent())?.trim?.() : x
    );
  }

  // Filter nil values
  if (Array.isArray(results)) {
    results = results.filter(x => x !== undefined && x !== null);
  }

  // Make unique
  if (Array.isArray(results) && obj.unique) {
    results = [...new Set(results)];
  }

  // Pick
  let pickIndex = undefined;
  if (Array.isArray(results)) {
    if (results.length === 1) {
      pickIndex = 0;
    } else if (obj.pick) {
      pickIndex = obj.pick(results);
    }
  }

  // Highlight matches
  const result = pickIndex !== undefined ? results[pickIndex] : results;
  if (has(results)) {
    if (obj.$) {
      await page.$eval(
        obj.$,
        (x, args) => {
          x.classList.add('catirpel-highlight');
          x.classList.add('selected');
          x.dataset.tippyContent = args.join('<br /><br />');
        },
        [obj.$, JSON.stringify(result), typeof result]
      );
    } else if (obj.$$) {
      await page.$$eval(
        obj.$$,
        (x, args) =>
          x.map((y, i) => {
            y.classList.add('catirpel-highlight');
            if (args[3] === undefined || i === args[3]) {
              y.classList.add('selected');
            }
            y.dataset.tippyContent = [args[0], args[1], args[2]].join(
              '<br /><br />'
            );
          }),
        [obj.$$, JSON.stringify(result), typeof result, pickIndex]
      );
    }
  }

  return result;
};

const select = async (page, stuff) => {
  const data = await Promise.all(
    Object.entries(stuff).map(async ([key, value]) => {
      let result;
      if (Array.isArray(value)) {
        result = (await Promise.all(value.map(x => evaluate(page, x)))).find(
          x => x !== undefined
        );
      } else {
        result = await evaluate(page, value);
      }
      return { [key]: result };
    })
  );

  return data.reduce((acc, x) => ({ ...acc, ...x }));
};

export default select;
